/*
  Eikon is an image similarity search engine based on wavelets.
  Copyright (C) 2002 Luke Francl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package org.francl.eikon.database;

import org.francl.eikon.wavelet.Signature;
import org.francl.eikon.wavelet.ColorChannelNotSupportedException;
import org.francl.eikon.util.ResourceId;
import org.francl.eikon.util.ImageMetaData;
import org.francl.eikon.util.SearchArray;

import java.io.Serializable;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;

/** The ImageMetaDataCache is a way to store information about ImageMetaData.
    It can be used to cache results from a relational or file database in
    memory, or it can be used to back a memory-based serialized database for
    storage. **/
public class ImageMetaDataCache implements Serializable {

    /** Contains all the ImageMetaData objects in the database. **/
    private HashMap cache = new HashMap();
    private SearchArrayCache searchArrays = new SearchArrayCache();

    public void addImage(ImageMetaData imd, Signature sig)
	throws ColorChannelNotSupportedException {
	synchronized(this) {
	    cache.put(imd.getResourceId(), imd);
	    searchArrays.add(imd.getResourceId(), sig);
	}
    }
    
    public void removeImage(ResourceId id) {
	synchronized(this) {
	    cache.remove(id);
	    searchArrays.remove(id);
	}
    }

    /** Updated the ImageMetaData object imd, but not the search arrays that
	imd belongs to. Use this to change the ImageMetaData object in some
	way, for example, updated its URL.
	@param imd the ImageMetaData object to update. ImageMetaData objects
	are compared by ResourceId, so it must have the same ResourceId as
	the object you want to replace.
	@throws ImageMetaDataNotFoundException if imd does not exist in the
	cache. This prevents the cache from getting in an inconsistent state.
	
    **/
    public void updateMetaData(ImageMetaData imd)
	throws ImageMetaDataNotFoundException {
	ResourceId id = imd.getResourceId();
	synchronized(cache) {
	    if(cache.containsKey(id) == false) {
		String m = "ImageMetaData " + imd + " was not found in " + 
		    " the database, so it could not be removed.";
		throw new ImageMetaDataNotFoundException(m);
	    } else {
		cache.put(imd.getResourceId(), imd);
	    }
	}
    }

    /** Returns the ImageMetaData object associated with
	the ResourceId id. If there is none, it returns null
    **/
    public ImageMetaData getImageMetaData(ResourceId id) {
	synchronized (cache) {
	    return (ImageMetaData)cache.get(id);
	}
    }

    /** Tells you if I have this one or not **/
    public boolean contains(ImageMetaData imd) {
	synchronized (cache) {
	    if (cache.containsKey(imd.getResourceId())) {
		return true;
	    } else {
		return false;
	    }
	}
    }
    
    public Iterator getAllImageMetaData() {
	synchronized(cache) {
	    Collection values = cache.values();
	    return values.iterator();
	}
    }

    /** Actually implementing this will increase my storage (and memory!)
	requirements dramatically...is in necessary?
    **/
    public Signature getSignature(ResourceId id) {
	throw new UnsupportedOperationException("Do I really need this?");
    }
    

    public Collection matches(int xPos, int yPos, String color, int value) {
	synchronized(searchArrays) {
	    Collection matches;
	    if (value == 1) {
		matches = searchArrays.matches(xPos, yPos, color, value);
	    } else if (value == -1) {
		matches = searchArrays.matches(xPos, yPos, color, value);
	    } else {
		String m = "Value was: " + value + ". It must be 1 or -1.";
		throw new IllegalArgumentException(m);
	    }
	    return matches;
	}

    }

    public Collection matches(int xPos, int yPos, char color, int value,
			      String basis, int width, int height) {
	throw new UnsupportedOperationException("Not done yet.");
    }
    
    /** Return a string telling me how many elements there are in this
	collection. 
    **/
    public String toString() {
	String s = "This ImageMetaDataCache contains ";
	s = s + cache.size() + " ImageMetaData elements";
	return s;
    }
}
