/*
  Eikon is an image similarity search engine based on wavelets.
  Copyright (C) 2002 Luke Francl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package org.francl.eikon.database;

import org.francl.eikon.util.SearchArray;
import org.francl.eikon.util.ImageMetaData;
import org.francl.eikon.util.ResourceId;
import org.francl.eikon.wavelet.Signature;
import org.francl.eikon.wavelet.ColorChannelNotSupportedException;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Collection;
import java.util.Iterator;

/** Holds multiple SearchArrays (six as the case may be), and provides 
    methods for loading entire ImageData objects into all the SearchArrays.
    It also has remedial support for removing objects from the SearchArrays.

    @author Luke Francl
**/

public class SearchArrayCache implements Serializable {

    public static final boolean DEBUG = false;
    private SearchArrayCollection searchArrays;

    public SearchArrayCache() {
	System.out.println("SearchArrayCache: creating new SearchArrayCache");
	searchArrays = new SearchArrayCollection();
    }

    /** Put the ResourceId in the appropriate bins of the SearchArrays.
	@param id the ResourceId of the image to be added
	@param sig the Signature with the data to be added
    **/
    public void add(ResourceId id, Signature sig)
	throws ColorChannelNotSupportedException {
	Iterator colors = sig.getColors();
	while (colors.hasNext()) {
	    String color = (String)colors.next();
	    int[][] curCoefs = sig.getCoeffients(color);
	    
	    for(int y=0; y<curCoefs.length; y++) {
		for(int x=0; x < (curCoefs[0]).length; x++) {
		    // Remember, if coeffient == 0, we don't want to add it
		    if (curCoefs[y][x] == 0) {
			continue;
		    }
		    if(DEBUG) {
			String msg = "SearchArrayCache.add: " +
			    "Adding " + id + " to (x=" + x +",y=" + y + "," +
			    "color=" + color +",sign=" +curCoefs[y][x] + ") " +
			    "bin";
			System.out.println(msg);
		    }

		    addCoeffient(curCoefs[y][x], id, x, y, color);
		}
	    }
	}
    }
    
    /** Remove a ResourceId from the search arrays. Each search array must
	be checked.
	@param id the ResourceId to be removed.
    **/
    public void remove(ResourceId id) {
	Iterator colors = searchArrays.getColors();
	while(colors.hasNext()) {
	    String color = (String)colors.next();
	    SearchArray pos =
		searchArrays.getSearchArray(color,
					    SearchArrayCollection.POSITIVE);
	    SearchArray neg =
		searchArrays.getSearchArray(color,
					    SearchArrayCollection.NEGATIVE);
	    pos.remove(id);
	    neg.remove(id);
	}
    }

    /** Adds a coeffient to the appropriate SearchArray. Creates the
	SearchArray if it doesn't exist yet.
	@param coeffient the value of the coeffient to be placed in the
	SearchArray. Must be -1, 0, or 1 (coeffients that equal zero will
	not be added to the SearchArray, but they are still legal arguments).
	@param id the ResourceId to be added. Must not be null.
	@param xPos the x position of coeffient
	@param yPos the y position of coeffient
    **/
    public void addCoeffient(int coeffient, ResourceId id, int xPos, int yPos,
			     String color) {
	if (id == null) {
	    throw new NullPointerException("id can't be null");
	}

	SearchArray pos =
	    searchArrays.getSearchArray(color, SearchArrayCollection.POSITIVE);
	SearchArray neg =
	    searchArrays.getSearchArray(color, SearchArrayCollection.NEGATIVE);
	

	if (coeffient == 1) {
	    pos.add(id, xPos, yPos);
	} else if (coeffient == -1) {
	    neg.add(id, xPos, yPos);
	} else {
	    String m = "coeffient ( " + coeffient + ") must be -1 or 1";
	    throw new IllegalArgumentException(m);
	}
    }

    /** Get all the matches at (x, y) for (color, value)
	@param x the x position of the search
	@param y the y position of the search
	@param color the color for the search. Must be one character long.
	@param value the value to match for the search. Must be -1 or 1.
    **/
    public Collection matches(int x, int y, String color, int value) {

	Collection matches;
	SearchArray sa;
	if (value == 1) {
	    sa = searchArrays.getSearchArray(color,
					     SearchArrayCollection.POSITIVE); 
	} else if (value == -1) {
	    sa = searchArrays.getSearchArray(color,
					     SearchArrayCollection.NEGATIVE);
	} else {
	    String m = "value (" + value + ") must be -1 or 1";
	    throw new IllegalArgumentException(m);
	}

	if (sa == null) {
	    String  m = "positiveMatches was called before that SearchArray " +
		"was initalized.";
	    throw new IllegalStateException(m);
	}
	
	matches = sa.matches(x, y);
	
	return matches;
    }
}
