package org.francl.eikon.image;

import org.francl.eikon.util.ResourceId;
import org.francl.eikon.wavelet.Signature;
import org.francl.eikon.util.ImageMetaData;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.net.URL;
import java.net.MalformedURLException;
import javax.imageio.*;
import javax.imageio.stream.*;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.awt.image.BufferedImage;
import java.util.Iterator;



/** ImageDownloader is the 'frontend' of the system. It retrieves 
    images from a URL and orchestrates processing that changes a raw
    image into a ImageData object (i.e., calculating the hash, resizing
    the image, translating to YIQ color space, performing wavelet 
    decomposition, etc.). I say it is a frontend, because it could be 
    swapped out for a class that reads images from a file with no ill
    effects for the rest of the system.
**/

public class ImageDownloader {

    /** Test method **/
    public static void main(String args[]) {
	if(args.length != 1) {
	    System.out.println("Usage: java ImageDownloader url");
	    System.exit(0);
	}
	
	try {
	    
	    System.out.println("Creating a new ImageDownloader object...");
	    ImageDownloader down = new ImageDownloader(args[0]);
	    ImageMetaData imd = down.getImageMetaData();
	    Signature sig = down.getSignature();
	    System.out.println("Here's the ImageMetaData object retrieved...");
	    System.out.print(imd);
	    System.out.println("Here's the Signature retrieved...");
	    System.out.println(sig);
	} catch (Exception e) {
	    e.printStackTrace(System.err);
	}
    }
    
    private ImageMetaData imd  = new ImageMetaData();
    private Signature sig;
    
    public ImageDownloader(String surl) throws MalformedURLException,
	IOException, NoSuchAlgorithmException, ImageTooSmallException, 
	UnsupportedImageTypeException {
	
	URL u = new URL(surl);
	
	imd.setURL(u);

	String contentType = (u.openConnection()).getContentType();
	if(contentType != null && !(contentType.equals("image/jpeg") ||
				    contentType.equals("image/jpg")  ||
				    contentType.equals("image/gif")  ||
				    contentType.equals("image/png"))) {
	    
	    String msg = "The image type was " + contentType + ". Only " +
		"GIF, JPEG, and PNG are supported at this time.";
	    throw new UnsupportedImageTypeException(msg);
	}

	if (contentType.equals("image/jpg")) {
	    contentType = "image/jpeg";
	}
	
	imd.setImageType(contentType);
	
	byte[] myImage = getImage(u);
	
	MessageDigest md = MessageDigest.getInstance("SHA");
	byte[] theDigest =  md.digest(myImage);
	ResourceId myID = new ResourceId(theDigest);
	
	imd.setResourceId(myID);
	
	BufferedImage bf = imageFromByte(myImage, contentType);
	
	imd.setWidth(bf.getWidth());
	imd.setHeight(bf.getHeight());

	// TODO: make this class image type agnostic. For right now
	// this should be OK.
	Image image = new YIQImage(bf);
	sig = image.getSignature();
    }
    
    /** Downloads the image from the URL and stores it in a byte array **/
    public byte[] getImage(URL url) throws IOException {
	// I want to use openStream to get a raw stream of bytes
	InputStream in = url.openStream();
	
	// download the image
	ByteArrayOutputStream bout = new ByteArrayOutputStream();
	int i;
	while((i = in.read()) != -1) {
	    bout.write(i);
	}
	byte[] theImage = bout.toByteArray();
	
	return theImage;
    }

    /** Converts a byte array into a BufferedImage object. **/
    public BufferedImage imageFromByte(byte[] image, String imageType) 
	throws IOException {
	
	// set up the image reader
	ByteArrayInputStream bin = new ByteArrayInputStream(image);
	Iterator readers = ImageIO.getImageReadersByMIMEType(imageType);
	ImageReader reader = (ImageReader)readers.next();
	ImageInputStream iis = ImageIO.createImageInputStream(bin);
	
	// attach it to the reader
	reader.setInput(iis, true);
	
	// put into a buffered image
	BufferedImage bf = reader.read(0);

	// hopefully, free up some memory
	iis.close();
	reader.dispose();

	return bf;
    }

    /** @return The ImageMetaData object created by invoking ImageDownloader's
	contstructor 
    **/
    public ImageMetaData getImageMetaData() {
	return imd;
    }

    public Signature getSignature() {
	return sig;
    }
}
