/*
  Eikon is an image similarity search engine based on wavelets.
  Copyright (C) 2002 Luke Francl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package org.francl.eikon.image;

import java.awt.image.BufferedImage;

import javax.imageio.*;

import org.francl.eikon.wavelet.Signature;
import org.francl.eikon.wavelet.YIQSignature;
import org.francl.eikon.wavelet.HaarWavelet;
import org.francl.eikon.wavelet.Wavelet;
import org.francl.eikon.wavelet.Truncator;
import org.francl.eikon.wavelet.Quantizor;
import org.francl.eikon.wavelet.ColorChannelNotSupportedException;

/** A YIQImage is a representation of a BufferedImage object which has
    been converted to the YIQ color space. Internally, this representation
    is stored as 3 two dimensional arrays (one each for the Y, I, and Q
    channels). The constructor for YIQ image handles the rescaling and
    conversion process. **/
public class YIQImage extends Image {
    public static final int NUM_LARGEST   =  60;
    public static final boolean DEBUG     = true;

    private static final double NO_VALUE  = -666.666;

    public static final String COLORSPACE = "YIQ";

    private double[][] Yelements;
    private double[][] Ielements;
    private double[][] Qelements;

    private double Yavg = NO_VALUE;
    private double Iavg = NO_VALUE;
    private double Qavg = NO_VALUE;

    private int height;
    private int width;
    
    private BufferedImage original;

    /** Constructor when you don't have an image to make a YIQImage out of,
	you just want to fill out its arrays, baby.
    **/
    public YIQImage() {}

    public YIQImage(BufferedImage image) throws ImageTooSmallException {
	width  = image.getWidth();
	height = image.getHeight();
	// rescale the image

	Converter convert = new Converter(image, YIQImage.COLORSPACE);
	Yelements = convert.getElements('Y');
	Ielements = convert.getElements('I');
	Qelements = convert.getElements('Q');
    }

    /** Constructor for testing purposes. You don't need to specify a
	BufferedImage, just give it the Y,I, and Q arrays.
	TODO: this should perform error checking for size.
    **/
    public YIQImage(double[][] Ychannel, double[][] Ichannel, 
		    double[][] Qchannel) {
	Yelements = Ychannel;
	Ielements = Ichannel;
	Qelements = Qchannel;
    }

    /** TODO: this should do checking on element size, yeah...**/
    public void setElements(char color, double[][] elements)
	throws ColorChannelNotSupportedException {
	switch(color) {
	case 'Y':
	    Yelements = elements;
	    break;
	case 'I':
	    Ielements = elements;
	    break;
	case 'Q':
	    Qelements = elements;
	    break;
	default:
	    String m = "This image type does not support color " + color;
	    throw new ColorChannelNotSupportedException(m);
	}
    }

    public String getColorSpace() {
	return COLORSPACE;
    }

    public double getAverage(char color)
	throws ColorChannelNotSupportedException {
	double avg;
	switch(color) {
	case 'Y':
	    avg = Yavg;
	    break;
	case 'I':
	    avg = Iavg;
	    break;
	case 'Q':
	    avg = Qavg;
	    break;
	default:
	    String m = "YIQImage doesn't support color " + color;
	    throw new ColorChannelNotSupportedException(m);
	}

	if(avg == NO_VALUE) {
	    String m = "Oops, you tried to get the average color " +
		"for an image that hasn't had it set yet. " + 
		"That probably means this image isn't being used to hold " +
		"decomposed data.";
	    throw new IllegalStateException(m);
	}
	return avg;
    }

    public void setAverage(char color, double avg)
	throws ColorChannelNotSupportedException {
	switch(color) {
	case 'Y':
	    Yavg = avg;
	    break;
	case 'I':
	    Iavg = avg;
	    break;
	case 'Q':
	    Qavg = avg;
	    break;
	default:
	    String m = "Blah, can't use color " + color;
	    throw new ColorChannelNotSupportedException(m);
	}
    }

    public double[][] getElements(char color)
	throws ColorChannelNotSupportedException {
	double[][] elements;
	switch (color) {
	case 'Y':
	    elements = Yelements;
	    break;
	case 'I':
	    elements = Ielements;
	    break;
	case 'Q':
	    elements = Qelements;
	    break;
	default:
	    throw new ColorChannelNotSupportedException();
	}

	return elements;
    }

    public int getHeight() {
	return height;
    }

    public int getWidth() {
	return width;
    }


    /** TODO: This needs to be configurable for wavelet type!
	Probably get from a a properties file. Eventually.
	@returns the Signature for this Image **/
    public Signature getSignature() {
	Wavelet wavelet = new HaarWavelet();
	Image decomposed = wavelet.decompose(this);
	
	Truncator trunc = new Truncator();
	Image truncated = trunc.truncate(decomposed, NUM_LARGEST);

	Quantizor quant = new Quantizor();
	Signature quantized = quant.quantize(truncated);

	return quantized;
    }
}
