/*
  Eikon is an image similarity search engine based on wavelets.
  Copyright (C) 2002 Luke Francl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package org.francl.eikon.query;
    
import org.francl.eikon.database.Database;
import org.francl.eikon.database.DatabaseFactory;
import org.francl.eikon.database.DatabaseImplNotFoundException;
import org.francl.eikon.util.ArrayListWrapper;
import org.francl.eikon.util.ImageMetaData;
import org.francl.eikon.util.ResourceId;
import org.francl.eikon.wavelet.Signature;
import org.francl.eikon.wavelet.YIQSignature;
import org.francl.eikon.wavelet.ColorChannelNotSupportedException;

import java.util.Iterator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Collections;
import java.util.Collection;
import java.util.List;
import java.util.ArrayList;

/** The Query class performs the scoring on a Signature object. 
    
    @author Luke Francl
**/
public class Query {

    public static final boolean DEBUG = false;
    
    private Map scores = Collections.synchronizedMap(new HashMap());
    private int results;
    private List scoredResourceIds;
    private Database db;
    
    public Query(Signature sig, int numResults)
	throws NotEnoughResultsException {
	if (numResults < 1) {
	    String m = "Number of results ( "+ numResults+") must be > 0";
	    throw new IllegalArgumentException();
	}

	System.out.println("Signature holds: " + sig.toString());
	db = DatabaseFactory.getDatabase();
	results = numResults;
	score(sig);
	scoredResourceIds = getTopScores(numResults);
    }

    public void score(Signature sig) throws ColorChannelNotSupportedException {
	Iterator colors = sig.getColors();
	if (!sig.getColorSpace().equals(YIQSignature.COLORSPACE)) {
	    // this is a colorSPACE not a channel...
	    throw new ColorChannelNotSupportedException("HACK!");
	}

	while(colors.hasNext()) {
	    String color = (String)colors.next();
	    System.out.println("avg for " + color + " is " +
			       sig.getAverage(color));
	    doScore(sig.getCoeffients(color), sig.getAverage(color),
		    color, sig.getWeights(color));
	}
    }

    /** Performs the scoring for a color channel.
	@param coeffients the truncated top n wavelet coeffient array
	@param channelAvg the average color of this channel for the
	query imagel
	@param channel the channel this is. Must be Y, I, or Q (for now).
	@param the weights to use for scoring. there must be 6 of them.
    **/
    public void doScore(int[][] coeffients, double channelAvg, String color,
			 double[] weights)
	throws DatabaseImplNotFoundException {

	if (weights.length != 6) {
	    String m = "Weight array wrong length (" + weights.length + ")";
	    throw new IllegalArgumentException(m);
	}

	setInitialScore(color, weights[0], channelAvg);
	for(int y=0; y < coeffients.length; y++) {
	    for(int x=0; x < (coeffients[0]).length; x++) {
		if (coeffients[y][x] == 0) {
		    continue;
		}
		Collection matches = db.matches(x, y, color, coeffients[y][x]);
		
		if(DEBUG) {
		    String msg = "matches for x=" + x +", y=" + y +
			", color=" + color + ",coeffient=" +
			coeffients[y][x] + ": " + matches;
		    System.out.println(msg);
		}
		
		Iterator it = matches.iterator();
		while(it.hasNext()) {
		    ResourceId id = (ResourceId)it.next();
		    Score myScore = (Score)scores.get(id);
		    double currentImageScore = myScore.getScore();
			
		    currentImageScore -= weights[bin(x,y)];
		    myScore.setScore(currentImageScore);
		    scores.put(id, myScore);
		    
		    if(DEBUG) {
			System.out.println("Calculating score");
			System.out.println("Score for ResourceId " + id + 
					   " set to " +currentImageScore);
		    }
		}
	    }
	}
    }

    /** Loops through all the images in the database and tags them with
	an inital score.
	WHY CAN'T I JUST DO THIS TO THE HIGH SCORING ONES AFTER IT'S DONE?
	TODO: Implement this alternative algorithm, above ^
	TODO: This method is called 3 times, and retrieves the ImageMetaData
	3 times. Instead, call it once and score each image 3 times!
	This is why there is a huge pause between invocations of this method!
    **/
    private void setInitialScore(String color, double weight,
				 double channelAvg) {
	Iterator images = db.getAllImageMetaData();
	if (!images.hasNext()) {
	    System.out.println("Fuck, there's no images!");
	}

	int i = 0;
	while(images.hasNext()) {
	    ImageMetaData currentImage = (ImageMetaData)images.next();
	    ResourceId myId = currentImage.getResourceId();

	    double targetAvg = currentImage.getAverage(color);
	    Score tmp = (Score) scores.get(myId);
	    double currentScore = 0;
	    if(tmp != null) {
		currentScore = tmp.getScore();
	    }
		       
	    double newScore = weight * Math.abs(targetAvg - channelAvg) +
		currentScore;

	    scores.put(myId, new Score(myId, newScore));
	    if (DEBUG) {
		String s = "Setting " + color + " score for " + i + "th" +
		    " image: " + currentImage + " to " + newScore +
		    " weight: " + weight + " target avg: " + targetAvg;
		System.out.println(s);
	    }
	    i++;
	}
	
	if(DEBUG) {
	    System.out.println("Scores are: " + scores);
	}

    }

    
    /** Returns a List of the ResourceIds with the top scores.
	@param numScores the number of top scores to return
	@returns a List of ResourceIds
    **/
    public List getTopScores(int numScores) throws NotEnoughResultsException {
	if(numScores <= 0) {
	    String m = "Number of scores requested must be greater than 0";
	    throw new IllegalArgumentException(m);
	}
	
	if(scores.isEmpty()) {
	    String m = "The scores list was empty";
	    throw new NotEnoughResultsException(m);
	} else {
	    ArrayListWrapper scoreList = 
		new ArrayListWrapper(scores.values());

	    if(scoreList.size() < numScores) {
		String m = "Total number of scores is " + scoreList.size() +
		    ". You asked for " + numScores;
		throw new NotEnoughResultsException(m);
	    }

	    Collections.sort(scoreList);
	    
	    if(DEBUG) {
		System.out.println("Here is the raw sorted list of scores...");
		System.out.println(scoreList);
	    }
	    

	    // LOWEST scores are the best matches
	    scoreList.removeRange(numScores, scoreList.size());
	    
	    if(DEBUG) {
		System.out.println("Here are the scores after truncation...");
		System.out.println(scoreList);
	    }
	    
	    // now create an ArrayList of *only* the ResourceIds
	    List resourceIdList = new ArrayList();
	    Iterator i = scoreList.iterator();
	    while( i.hasNext()) {
		resourceIdList.add(((Score)i.next()).getResourceId());
	    }
	    
	    return resourceIdList;
	} 
    }
    
    /** Groups different coeffients into a small number of bins 
	@param x the x postion of the coeffient
	@param y the y position of the coeffient
	@see Wavelets for Computer Graphics, page 83
    **/
    private int bin(int x, int y) {
	double ln_2 = Math.log(2);

	// TODO: 5d should be a constant.
	double bin = Math.min(Math.max(Math.floor(Math.log(x)/ln_2), 
				       Math.floor(Math.log(y)/ln_2)), 5d);
	
	// so, why exactly does Math.floor return a double?
	Double d = new Double(bin);
	return d.intValue();
    }

    public ResultSet getResultSet(ResourceId queryImage) {
	return new ResultSet(scoredResourceIds, queryImage);
    }

    public ResultSet getResultSet() {
	return new ResultSet(scoredResourceIds);
    }
}
